import argparse
import io
import re
import sys
import xml.etree.ElementTree as ET
from html import unescape
from pathlib import Path
from urllib.parse import quote
from urllib.request import Request, urlopen
from pprint import pformat

from lawtext import parse, render, util



def main(infile, intype=None, outfile=None, outtype=None):
    if not intype:
        if infile.startswith('http:') or infile.startswith('https:'):
            intype = 'web_xml'
        else:
            infilename = Path(infile).name
            if infile.startswith('http:') or infile.startswith('https:'):
                intype = 'web_xml'
            elif infilename.endswith('.xml'):
                intype = 'xml'
            elif infilename.endswith('.law.txt'):
                intype = 'lawtext'
            else:
                intype = 'lawtext'

    if not outtype:
        outfilename = Path(outfile).name if outfile else ''
        if outfilename.endswith('.xml'):
            outtype = 'xml'
        elif outfilename.endswith('.html'):
            outtype = 'html'
        elif outfilename.endswith('.law.txt'):
            outtype = 'lawtext'
        elif outfilename.endswith('.lexed.py'):
            outtype = 'lexed'
        elif outfilename.endswith('.docx'):
            outtype = 'docx'
        elif intype == 'xml':
            outtype = 'lawtext'
        elif intype == 'lawtext':
            outtype = 'xml'
        else:
            outtype = 'xml'

    if outtype == 'lexed' and intype != 'lawtext':
        raise ValueError('OUTTYPE=lexed can be used only with INTYPE=lawtext.')

    if intype == 'web_xml':
        def q(string):
            return quote(string.group(0))
        url = re.sub(r'[^ \x00-\x7f]+', q, infile)
        req = Request(url)
        req.add_header('User-Agent', 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/62.0.3202.94 Safari/537.36')
        with urlopen(req) as f:
            response = f.read().decode('utf-8')
        xml_el = ET.fromstring(response)
        xml = unescape(xml_el.find('.//LawFullText').text)
        law = util.xml_to_json(xml)
    elif intype == 'xml':
        xml = Path(infile).read_text(encoding='utf-8')
        law = util.xml_to_json(xml)
    elif intype == 'lawtext':
        lawtext = Path(infile).read_text(encoding='utf-8')
        if outtype == 'lexed':
            lexed = parse.lex(lawtext.splitlines())
        else:
            law = parse.parse_lawtext(lawtext)

    if outtype == 'docx':
        outbytes = render.render_docx(law)
        if outfile:
            Path(outfile).write_bytes(outbytes)
        else:
            sys.stdout.buffer.write(outbytes)
    else:
        if outtype == 'lawtext':
            outtext = render.render_lawtext(law)
        elif outtype == 'xml':
            outtext = render.render_xml(law)
        elif outtype == 'html':
            outtext = render.render_html(law)
        elif outtype == 'htmlfragment':
            outtext = render.render_htmlfragment(law)
        elif outtype == 'htmlfragment':
            outtext = render.render_htmlfragment(xml)
        elif outtype == 'lexed':
            outtext = pformat(lexed, indent=2)

        if outfile:
            Path(outfile).write_text(outtext, encoding='utf-8')
        else:
            print(outtext)



if __name__ == '__main__':

    try:
        sys.stdout = io.TextIOWrapper(
            sys.stdout.buffer,
            encoding=sys.stdout.encoding,
            errors='backslashreplace',
            line_buffering=sys.stdout.line_buffering,
        )
    except:
        pass

    argparser = argparse.ArgumentParser()
    argparser.add_argument('infile')
    argparser.add_argument(
        '-it', '--intype',
        choices=['lawtext', 'xml', 'web_xml'],
    )
    argparser.add_argument('outfile', nargs='?')
    argparser.add_argument(
        '-ot', '--outtype',
        choices=['lawtext', 'xml', 'html', 'htmlfragment', 'docx', 'lexed'],
    )
    args = argparser.parse_args()
    main(**vars(args))



import argparse
import locale
import os
import shutil
import subprocess
import sys
from distutils.dir_util import copy_tree
from pathlib import Path
from shutil import rmtree

from jinja2 import Environment, FileSystemLoader

env = Environment(loader=FileSystemLoader(str(Path(__file__).parent.resolve() / 'htmlapp_templates'), encoding='utf8'))

def write_text(path, text):
    path.parent.mkdir(exist_ok=True, parents=True)
    path.touch()
    path.write_text(text, encoding='utf-8')

def main(outdir, preserve_compiled_files, ie=False):
    outdir = Path(outdir).resolve()
    if outdir.exists():
        rmtree(str(outdir))

    copy_tree(
        str(Path(__file__).parent.resolve() / 'htmlapp_templates'),
        str(outdir),
    )

    shutil.copyfile(
        str(Path(__file__).parent.resolve() / 'static' / 'law.css'),
        str(Path(outdir / 'src' / 'law.css')),
    )

    templates_js_path = Path(outdir / 'src' / 'templates.js')
    if not preserve_compiled_files or not templates_js_path.exists():
        outtext = subprocess.check_output([
            'nunjucks-precompile',
            '--include', '.+',
            str(Path(__file__).parent.resolve() / 'templates'),
        ], shell=True).decode(encoding='utf-8')
        write_text(templates_js_path, outtext)

    import_parse_decorate_js_path = Path(outdir / 'src' / 'parse_decorate.js')
    if not preserve_compiled_files or not import_parse_decorate_js_path.exists():
        subprocess.check_call([
            'transcrypt',
            '--build',
            '--esv', '5',
            '--fcall',
            '--dextex',
            '--nomin',
            '--xpath', str(Path(__file__).parents[1].resolve()),
            str(Path(__file__).parent.resolve() / '_parse_decorate.py'),
        ])
        outtext = (Path(__file__).parent.resolve() / '__javascript__' / '_parse_decorate.js').read_text(encoding='utf-8')
        outtext = outtext.replace('error.args', 'error.__args__')
        outtext = outtext.replace('join (self.args)', 'join (self.__args__)')
        outtext = outtext.replace(', __kwargtrans__ (kwargs)', ', __kwargtrans__ ([].slice.apply (arguments).slice (2))')
        write_text(import_parse_decorate_js_path, outtext)

    if ie:

        pd_path = Path(outdir / 'src' / 'parse_decorate.js')
        pd_text = pd_path.read_text(encoding='utf-8')
        pd_text = pd_text.replace(
            'new RegExp(pattern, jsFlags)',
            'new RegExp(pattern, jsFlags.replace("u", ""))',
        )
        pd_path.write_text(pd_text, encoding='utf-8')

        es6_files = list(Path(outdir / 'src').glob('*.js'))
        es6_files += list(Path(outdir / 'templates').glob('*.html'))
        for path in es6_files:
            print(f'Converting {path.name} for IE')
            out_path = path.parent / ('ie_' + path.name)
            subprocess.check_call([
                'babel',
                '--no-babelrc',
                str(path.resolve()),
                '--out-file', str(out_path.resolve()),
                '--presets', 'es2015',
            ], shell=True)



if __name__ == '__main__':
    argparser = argparse.ArgumentParser()
    argparser.add_argument('outdir')
    argparser.add_argument('-p', '--preserve-compiled-files', action='store_true')
    argparser.add_argument('-ie', action='store_true')
    args = argparser.parse_args()
    main(**vars(args))



import re

DEFAULT_INDENT = '　'

class LawtextParseError(Exception):
    def __init__(self, message, lineno, lines):
        super().__init__(message)
        self.message = message
        self.lineno = lineno
        self.lines = lines
    def __str__(self):
        lines = [
            f'{ "==>" if lineno == self.lineno else "   " }{lineno:5}|{self.lines[lineno]}\n'
            for lineno in range(
                max(0, self.lineno - 2),
                min(len(self.lines), self.lineno + 2 + 1),
            )
        ]
        lines_str = ''.join(lines)
        return self.message + '\n\n' + lines_str

class LexerError(Exception):
    def __init__(self, message, lineno):
        super().__init__(message)
        self.message = message
        self.lineno = lineno

class LexerInternalError(Exception):
    def __init__(self, message):
        super().__init__(message)
        self.message = message

class ParserError(Exception):
    def __init__(self, message, lineno):
        super().__init__(message)
        self.message = message
        self.lineno = lineno


re_DETECT_INDENT = re.compile(r'^(?P<indent>[ \t\f\v]+)\S.*$')

def detect_indent(lines):
    indents = set()
    for line in lines:
        match = re_DETECT_INDENT.match(line)
        if match:
            indents.add(match.group(1))

    if not len(indents):
        return DEFAULT_INDENT

    indents = sorted(indents, key=len)
    return indents[0]

def indent_level(indents, indent):
    s, n = re.subn(indent, '', indents)
    if s:
        indent_description = indent
        indent_description = re.sub(' ', '<半角スペース>', indent_description)
        indent_description = re.sub('　', '<全角スペース>', indent_description)
        indent_description = re.sub('	', '<tab>', indent_description)
        raise LexerInternalError(f'インデントの文字が整っていません。なお、この文書で用いられているインデント文字は、「{indent}」＝「{indent_description}」であると推測されました。')
    return n

re_TC1stLine   = re.compile(r'^(?P<indents>\s*)\* - (?P<tcbody>.*)$')
re_TCnthLine   = re.compile(r'^(?P<indents>\s*)  - (?P<tcbody>.*)$')
re_FigLine   = re.compile(r'^(?P<indents>\s*)\.\.(?:\s+)figure::(?:\s+)(?P<figbody>.*)$')
re_DefaultLine = re.compile(r'^(?P<indents>\s*)(?P<linebody>\S.*)$')
re_BlankLine   = re.compile(r'^\s*$')

def lex_line(line, indent):

    match = re_TC1stLine.match(line)
    if match:
        return [
            'TC1stLine',
            indent_level(match.group(1), indent),
            lex_tcbody(match.group(2)),
        ]

    match = re_TCnthLine.match(line)
    if match:
        return [
            'TCnthLine',
            indent_level(match.group(1), indent),
            lex_tcbody(match.group(2)),
        ]

    match = re_FigLine.match(line)
    if match:
        return [
            'FigLine',
            indent_level(match.group(1), indent),
            {
                'body': match.group(2),
            }
        ]

    match = re_DefaultLine.match(line)
    if match:
        return [
            'DefaultLine',
            indent_level(match.group(1), indent),
            {
                'body': match.group(2),
            }
        ]

    match = re_BlankLine.match(line)
    if match:
        return [
            'BlankLine',
            0,
            {
                'body': '',
            }
        ]

    raise LexerInternalError('行の構造が判別できません。')

re_Arg = re.compile(r'^\[(?P<name>\S+?)="?(?P<value>\S+?)"?\]')

def lex_tcbody(tcbody):
    attr = {}
    body = tcbody
    while True:
        match = re_Arg.match(body)
        if not match:
            break
        attr[match.group(1)] = match.group(2)
        body = re_Arg.sub('', body, count=1)
    return {
        'attr': attr,
        'body': body,
    }

def lex(lines):
    indent = detect_indent(lines)
    ret = []
    for lineno, line in enumerate(lines):
        try:
            ret.append(lex_line(line, indent))
        except LexerInternalError as error:
            raise LexerError(error.message, lineno)

    return ret

PARENTHESIS_PAIRS = [
    (re.compile(r'[(（]'), re.compile(r'[)）]')),
    (re.compile(r'「'),    re.compile(r'」')),
]
re_FORCE_EXIT_PARENTHESIS = re.compile(r'」')
SENTENCE_DELIMITERS = [
    re.compile(r'。'),
]

def skip_parenthesis(text, pos):
    re_sp, re_ep = None, None
    for re_pair in PARENTHESIS_PAIRS:
        if re_pair[0].match(text[pos]):
            re_sp, re_ep = re_pair
            break
    if re_sp and re_ep:
        pos += 1
        while pos < len(text):
            old_pos = pos
            pos = skip_parenthesis(text, pos)
            if pos != old_pos:
                continue
            if re_ep.match(text[pos]):
                pos += 1
                break
            if re_FORCE_EXIT_PARENTHESIS.match(text[pos]):
                break
            pos += 1
    return pos

def split_sentences(text):
    sentences = []
    start_pos = 0
    pos = 0

    while pos < len(text):
        old_pos = pos
        pos = skip_parenthesis(text, pos)
        if pos != old_pos:
            continue
        for sentence_delimiter in SENTENCE_DELIMITERS:
            if sentence_delimiter.match(text[pos]):
                pos += 1
                sentences.append(text[start_pos:pos])
                start_pos = pos
                continue
        pos += 1
    if start_pos < len(text):
        sentences.append(text[start_pos:])

    return sentences

def split(text):
    split_text = []
    start_pos = 0
    pos = 0

    while pos < len(text):
        old_pos = pos
        pos = skip_parenthesis(text, pos)
        if pos != old_pos:
            continue
        if not text[pos].strip():
            split_text.append(text[start_pos:pos])
            pos += 1
            start_pos = pos
        pos += 1
    if start_pos < len(text):
        split_text.append(text[start_pos:])

    return split_text

def get_sentences(text):
    els = []
    text_split = split(text)

    if len(text_split) >= 2:
        for column in text_split:
            els.append({
                'tag': 'Column',
                'attr': {},
                'children': [{
                    'tag': 'Sentence',
                    'attr': {},
                    'children': [sentence],
                } for sentence in split_sentences(column)],
            })
    else:
        els.extend([{
            'tag': 'Sentence',
            'attr': {},
            'children': [sentence],
        } for sentence in split_sentences(text)])

    return els

re_LawNum            = re.compile(r'^[(（](?P<body>.+?)[）)](?:\s*)$')
re_ArticleRange      = re.compile(r'(?P<body>[(（].+?[）)])(?:\s*)$')
re_ArticleCaption    = re.compile(r'^[(（](?P<body>.+?)[）)](?:\s*)$')
re_ArticleBody       = re.compile(r'^(?P<title>\S+)\s+(?P<body>\S.*)(?:\s*)$')
re_ParagraphCaption  = re.compile(r'^[(（](?P<body>.+?)[）)](?:\s*)$')
re_ParagraphItemBody = re.compile(r'^(?P<title>\S+)\s+(?P<body>\S.*)(?:\s*)$')
re_SupplProvisionLabel = re.compile(r'^(?P<title>.+?)\s*(?:[(（](?P<amend_law_num>.+?)[）)])?(?:\s+(?P<extract>抄?))?(?:\s*)$')
re_AppdxTableLabel = re.compile(r'^(?P<title>.+?)\s*(?:(?P<related_article_num>[(（].+?[）)]))?(?:\s*)$')
re_AppdxStyleLabel = re.compile(r'^(?P<title>.+?)\s*(?:(?P<related_article_num>[(（].+?[）)]))?(?:\s*)$')

class Parser:
    def __init__(self, lexed_lines, lineno=0):
        self.lexed_lines = lexed_lines
        self.lineno = lineno
    def get(self, lineno):
        if 0 <= lineno < len(self.lexed_lines):
            return self.lexed_lines[lineno]
        else:
            return ('BlankLine', 0, {})
    def continuing(self):
        return self.lineno < len(self.lexed_lines)
    def forward(self):
        self.lineno += 1
    def here(self):
        return self.get(self.lineno)
    def prev(self):
        return self.get(self.lineno - 1)
    def next(self):
        return self.get(self.lineno + 1)



    def skip_blank_lines(self):
        while self.continuing():
            line_type, indent, data = self.here()
            if line_type == 'BlankLine':
                self.forward()
            else:
                break

    def process_title(self):
        self.skip_blank_lines()
        law_title = None
        law_num = None

        line_type, indent, data = self.here()
        next_line_type, next_indent, next_data = self.next()
        body_split = split(data['body'])

        if (
            line_type == 'DefaultLine' and
            indent == 0 and
            len(body_split) == 1 and
            (
                next_line_type == 'BlankLine' or

                # LawNum
                next_line_type == 'DefaultLine' and
                next_indent == 0
            )
        ):

            law_title = {
                'tag': 'LawTitle',
                'attr': {},
                'children': [data['body']],
            }
            self.forward()

            if self.continuing():
                line_type, indent, data = self.here()
                if line_type == 'DefaultLine':
                    match = re_LawNum.match(data['body'])
                    if not match:
                        ParserError('法令番号は括弧（　）で囲ってください。', lineno)
                    law_num = {
                        'tag': 'LawNum',
                        'attr': {},
                        'children': [match.group(1)],
                    }
                    self.forward()

        return law_title, law_num

    def process_enact_statement(self):
        self.skip_blank_lines()
        enact_statement = None

        line_type, indent, data = self.here()
        next_line_type, next_indent, next_data = self.next()
        body_split = split(data['body'])

        if (
            indent == 0 and
            len(body_split) == 1 and

            next_line_type == 'BlankLine'
        ):

            enact_statement = {
                'tag': 'EnactStatement',
                'attr': {},
                'children': [data['body']],
            }
            self.forward()

        return enact_statement


    def process_toc_group(self):
        self.skip_blank_lines()
        toc_group = None

        line_type, indent, data = self.here()
        next_line_type, next_indent, next_data = self.next()
        body_split = re.split(r'\s+', re.split('の', data['body'])[0])

        if (
            line_type == 'DefaultLine'
        ):

            toc_group_children = []

            toc_group_char = body_split[0][len(body_split[0]) - 1]

            title = re_ArticleRange.sub('', data['body'])
            title_tag = {
                '編': 'PartTitle', '章': 'ChapterTitle', '節': 'SectionTitle',
                '款': 'SubsectionTitle', '目': 'DivisionTitle', '条': 'ArticleTitle',
                '則': 'SupplProvisionLabel'
            }[toc_group_char]
            toc_group_children.append({
                'tag': title_tag,
                'attr': {},
                'children': [title],
            })

            match = re_ArticleRange.search(data['body'])
            if match:
                toc_group_children.append({
                    'tag': 'ArticleRange',
                    'attr': {},
                    'children': [match.group(1)],
                })

            self.forward()

            toc_base_indent = indent

            while self.continuing():
                line_type, indent, data = self.here()
                if line_type == 'DefaultLine':
                    if indent <= toc_base_indent:
                        break
                    elif indent != toc_base_indent + 1:
                        ParserError('目次の内容のインデントが整っていません。', lineno)
                    sub_toc_group = self.process_toc_group()
                    if sub_toc_group:
                        toc_group_children.append(sub_toc_group)
                    else:
                        break
                else:
                    break

            toc_tag = {
                '編': 'TOCPart', '章': 'TOCChapter', '節': 'TOCSection',
                '款': 'TOCSubsection', '目': 'TOCDivision', '条': 'TOCArticle',
                '則': 'TOCSupplProvision'
            }[toc_group_char]

            toc_group = {
                'tag': toc_tag,
                'attr': {},
                'children': toc_group_children,
            }

        return toc_group

    def process_toc(self):
        self.skip_blank_lines()
        toc = None

        line_type, indent, data = self.here()
        next_line_type, next_indent, next_data = self.next()
        body_split = split(data['body'])

        if (
            indent == 0 and
            len(body_split) == 1 and

            next_line_type == 'DefaultLine'
        ):
            toc_children = []
            toc_children.append({
                'tag': 'TOCLabel',
                'attr': {},
                'children': [data['body']],
            })
            self.forward()

            toc_base_indent = indent

            while self.continuing():
                line_type, indent, data = self.here()
                if line_type == 'DefaultLine':
                    if indent != toc_base_indent + 1:
                        ParserError('目次の内容のインデントが整っていません。', lineno)
                    toc_group = self.process_toc_group()
                    if toc_group:
                        toc_children.append(toc_group)
                    else:
                        break
                else:
                    break

            toc = {
                'tag': 'TOC',
                'attr': {},
                'children': toc_children,
            }

        return toc

    def process_main_provision(self):
        self.skip_blank_lines()
        main_provision = None
        main_provision_children = []

        while self.continuing():
            article_group = self.process_article_group()
            if article_group:
                main_provision_children.append(article_group)
            else:
                articles = self.process_articles()
                if len(articles):
                    main_provision_children.extend(articles)
                else:
                    break

        if len(main_provision_children):
            main_provision = {
                'tag': 'MainProvision',
                'attr': {},
                'children': main_provision_children,
            }

        return main_provision

    def process_article_group(self, parent_group_tag=''):
        self.skip_blank_lines()
        article_group = None

        line_type, indent, data = self.here()
        next_line_type, next_indent, next_data = self.next()
        body_split = re.split(r'\s+', re.split('の', data['body'])[0])
        article_group_char = body_split[0][len(body_split[0]) - 1]

        article_group_tag_rank = {
            '': 0, 'Part': 1, 'Chapter': 2, 'Section': 3,
            'Subsection': 4, 'Division': 5,
        }
        article_group_tags = {
            '編': 'Part', '章': 'Chapter', '節': 'Section',
            '款': 'Subsection', '目': 'Division',
        }

        if (
            indent != 0 and
            article_group_char in article_group_tags and
            article_group_tag_rank[article_group_tags[article_group_char]]
                > article_group_tag_rank[parent_group_tag] and

            next_line_type == 'BlankLine'
        ):
            article_group_tag = article_group_tags[article_group_char]
            article_group_title_tag = {
                '編': 'PartTitle', '章': 'ChapterTitle', '節': 'SectionTitle',
                '款': 'SubsectionTitle', '目': 'DivisionTitle', '条': 'ArticleTitle',
            }[article_group_char]

            article_group_children = []
            article_group_children.append({
                'tag': article_group_title_tag,
                'attr': {},
                'children': [data['body']],
            })
            self.forward()
            self.skip_blank_lines()

            while self.continuing():
                line_type, indent, data = self.here()

                sub_article_group = self.process_article_group(article_group_tag)
                if sub_article_group:
                    article_group_children.append(sub_article_group)
                else:
                    articles = self.process_articles()
                    if len(articles):
                        article_group_children.extend(articles)
                    else:
                        break

            if len(article_group_children):
                article_group = {
                    'tag': article_group_tag,
                    'attr': {},
                    'children': article_group_children,
                }

        return article_group

    def process_articles(self):
        self.skip_blank_lines()
        articles = []

        while self.continuing():
            article = self.process_article()
            if article:
                articles.append(article)
            else:
                break

        return articles

    def process_article(self):
        self.skip_blank_lines()
        article = None
        article_children = []

        line_type, indent, data = self.here()
        body_split = split(data['body']) if len(data) else ''
        next_line_type, next_indent, next_data = self.next()
        next_body_split = split(next_data['body']) if len(next_data) else ''
        next2_line_type, next2_indent, next2_data = self.get(self.lineno + 2)

        if (
            line_type == 'DefaultLine' and
            indent == 1 and
            re_ArticleCaption.match(data['body']) and

            next_line_type == 'DefaultLine' and
            next_indent == 0 and
            not (next2_data['body'].startswith('別表') if len(next2_data) else False) and
            (
                re_ArticleBody.match(next_data['body']) or
                (
                    # No Body
                    len(next_body_split) == 1 and
                    next2_line_type == 'DefaultLine' and
                    next2_indent == 0 and
                    re_ParagraphItemBody.match(next2_data['body'])
                )
            )

        ):
            article_children.append({
                'tag': 'ArticleCaption',
                'attr': {},
                'children': [data['body']],
            })
            self.forward()
            # No skip blank line

            line_type, indent, data = self.here()
            body_split = split(data['body']) if len(data) else ''
            next_line_type, next_indent, next_data = self.next()

        match = re_ArticleBody.match(data['body'])
        if (
            line_type == 'DefaultLine' and
            indent == 0 and
            not data['body'].startswith('別表') and
            (
                match or
                (
                    # No Body
                    len(body_split) == 1 and
                    next_line_type == 'DefaultLine' and
                    next_indent == 0 and
                    re_ParagraphItemBody.match(next_data['body'])
                )
            )
        ):
            if match:
                article_title, _ = match.group(1), match.group(2)
            else:
                article_title = data['body']

            article_children.append({
                'tag': 'ArticleTitle',
                'attr': {},
                'children': [article_title],
            })

            first_paragraph_in_article = True

            if not match:
                article_children.append({
                    'tag': 'Paragraph',
                    'attr': {},
                    'children': [
                        {
                            'tag': 'ParagraphNum',
                            'attr': {},
                            'children': [],
                        },
                        {
                            'tag': 'ParagraphSentence',
                            'attr': {},
                            'children': [{
                                'tag': 'Sentence',
                                'attr': {},
                                'children': [],
                            }],
                        },
                    ],
                })
                self.forward()
                first_paragraph_in_article = False

            paragraph_items = self.process_paragraph_items(indent, first_paragraph_in_article)
            if len(paragraph_items):
                article_children.extend(paragraph_items)

            article = {
                'tag': 'Article',
                'attr': {},
                'children': article_children,
            }

        return article

    def process_paragraph_items(self, current_indent=0, in_article=False):
        self.skip_blank_lines()
        paragraph_items = []

        first_paragraph_in_article = in_article

        while self.continuing():
            paragraph_item = self.process_paragraph_item(current_indent,  first_paragraph_in_article)
            if paragraph_item:
                paragraph_items.append(paragraph_item)
                first_paragraph_in_article = False
            else:
                break

        return paragraph_items

    def process_paragraph_item(self, current_indent, first_paragraph_in_article=False):
        self.skip_blank_lines()
        paragraph_item = None
        paragraph_item_children = []

        line_type, indent, data = self.here()
        next_line_type, next_indent, next_data = self.next()
        next_match = re_ParagraphItemBody.match(next_data['body']) if len(next_data) else None

        if (
            not first_paragraph_in_article and

            line_type == 'DefaultLine' and
            indent == current_indent + 1 and
            re_ParagraphCaption.match(data['body']) and

            next_line_type == 'DefaultLine' and
            next_indent == current_indent and
            next_match and
            '条' not in next_match.group(1) and
            '附' not in next_match.group(1) and
            '付' not in next_match.group(1) and
            not next_match.group(1).startswith('別表')
        ):
            paragraph_item_children.append({
                'tag': 'ParagraphCaption',
                'attr': {},
                'children': [data['body']],
            })
            self.forward()
            # No skip blank line

            line_type, indent, data = self.here()
            next_line_type, next_indent, next_data = self.next()

        paragraph_item_tags = {
             0: 'Paragraph', 1: 'Item',
             2: 'Subitem1',  3: 'Subitem2',  4: 'Subitem3',
             5: 'Subitem4',  6: 'Subitem5',  7: 'Subitem6',
             8: 'Subitem7',  9: 'Subitem8', 10: 'Subitem9',
            11: 'Subitem10',
        }
        paragraph_item_title_tags = {
             0: 'ParagraphNum',  1: 'ItemTitle',
             2: 'Subitem1Title', 3: 'Subitem2Title', 4: 'Subitem3Title',
             5: 'Subitem4Title', 6: 'Subitem5Title', 7: 'Subitem6Title',
             8: 'Subitem7Title', 9: 'Subitem8Title', 10: 'Subitem9Title',
            11: 'Subitem10Title',
        }
        paragraph_item_sentence_tags = {
             0: 'ParagraphSentence',  1: 'ItemSentence',
             2: 'Subitem1Sentence', 3: 'Subitem2Sentence', 4: 'Subitem3Sentence',
             5: 'Subitem4Sentence', 6: 'Subitem5Sentence', 7: 'Subitem6Sentence',
             8: 'Subitem7Sentence', 9: 'Subitem8Sentence', 10: 'Subitem9Sentence',
            11: 'Subitem10Sentence',
        }

        match = re_ParagraphItemBody.match(data['body'])
        if (
            line_type == 'DefaultLine' and
            indent == current_indent and
            (
                (
                    match and
                    '附' not in match.group(1) and
                    '付' not in match.group(1) and
                    '備考' not in match.group(1) and
                    (
                        first_paragraph_in_article or
                        (
                            '条' not in match.group(1) and
                            not match.group(1).startswith('別表')
                        )
                    )
                ) or
                (
                    indent == 0 and
                    len(split(data['body'])) == 1 and
                    not data['body'].startswith('別表')
                )
            )
        ):

            if first_paragraph_in_article or not match:
                paragraph_item_children.append({
                    'tag': 'ParagraphNum',
                    'attr': {},
                    'children': [],
                })
            else:
                paragraph_item_title = match.group(1)
                paragraph_item_title_tag = paragraph_item_title_tags[current_indent]
                paragraph_item_children.append({
                    'tag': paragraph_item_title_tag,
                    'attr': {},
                    'children': [paragraph_item_title],
                })
            paragraph_item_sentence_children = get_sentences(match.group(2) if match else data['body'])
            paragraph_item_sentence_tag = paragraph_item_sentence_tags[current_indent]
            paragraph_item_children.append({
                'tag': paragraph_item_sentence_tag,
                'attr': {},
                'children': paragraph_item_sentence_children,
            })
            self.forward()

            paragraph_item_tag = paragraph_item_tags[current_indent]
            sub_paragraph_items = self.process_paragraph_items(current_indent + 1)
            if len(sub_paragraph_items):
                paragraph_item_children.extend(sub_paragraph_items)

            table_struct = self.process_table_struct(current_indent + 1)
            if table_struct:
                paragraph_item_children.append(table_struct)

            list_ = self.process_list(current_indent + 2)
            if list_:
                paragraph_item_children.append(list_)

            paragraph_item = {
                'tag': paragraph_item_tag,
                'attr': {},
                'children': paragraph_item_children,
            }

        return paragraph_item

    def process_table_struct(self, current_indent):
        self.skip_blank_lines()
        table_struct = None
        table_children = []
        table_struct_children = []

        remarks = self.process_remarks(current_indent)
        if remarks:
            table_struct_children.append(remarks)

        line_type, indent, data = self.here()
        next_line_type, next_indent, next_data = self.next()

        if (
            line_type == 'TC1stLine' and
            indent == current_indent
        ):
            table_rows = self.process_table_rows(current_indent)
            if len(table_rows):
                table_children.extend(table_rows)
                table = {
                    'tag': 'Table',
                    'attr': {},
                    'children': table_children,
                }
                table_struct_children.append(table)

        if len(table_struct_children):
            table_struct = {
                'tag': 'TableStruct',
                'attr': {},
                'children': table_struct_children,
            }

        return table_struct

    def process_table_rows(self, current_indent):
        table_rows = []

        while self.continuing():
            table_row = self.process_table_row(current_indent)
            if table_row:
                table_rows.append(table_row)
            else:
                break

        return table_rows

    def process_table_row(self, current_indent):
        self.skip_blank_lines()
        table_row = None
        table_row_children = []

        line_type, indent, data = self.here()
        next_line_type, next_indent, next_data = self.next()

        if (
            line_type == 'TC1stLine' and
            indent == current_indent
        ):
            table_columns = self.process_table_columns(current_indent)
            if len(table_columns):
                table_row_children.extend(table_columns)
                table_row = {
                    'tag': 'TableRow',
                    'attr': {},
                    'children': table_row_children,
                }

        return table_row

    def process_table_columns(self, current_indent):
        table_columns = []

        first_column_processed = False

        while self.continuing():
            line_type, indent, data = self.here()
            if (
                (
                    not first_column_processed and
                    line_type == 'TC1stLine'
                ) or (
                    first_column_processed and
                    line_type == 'TCnthLine'
                )
            ):
                table_column = self.process_table_column(current_indent)
                if table_column:
                    table_columns.append(table_column)
                    if not first_column_processed:
                        first_column_processed = True
                else:
                    break
            else:
                break

        return table_columns

    def process_table_column(self, current_indent):
        self.skip_blank_lines()
        table_column = None
        table_column_children = []

        line_type, indent, data = self.here()
        next_line_type, next_indent, next_data = self.next()

        if (
            line_type in ('TC1stLine', 'TCnthLine') and
            indent == current_indent
        ):
            table_column_children.append({
                'tag': 'Sentence',
                'attr': {},
                'children': [data['body']],
            })
            self.forward()

            table_column_attr = data.get('attr', {})

            table_column_base_indent = indent

            while self.continuing():
                line_type, indent, data = self.here()
                next_line_type, next_indent, next_data = self.next()

                if (
                    line_type == 'DefaultLine' and
                    indent == current_indent + 2
                ):
                    table_column_children.append({
                        'tag': 'Sentence',
                        'attr': {},
                        'children': [data['body']],
                    })
                    self.forward()
                else:
                    break

            table_column = {
                'tag': 'TableColumn',
                'attr': table_column_attr,
                'children': table_column_children,
            }

        return table_column

    def process_fig(self, current_indent):
        self.skip_blank_lines()
        fig = None

        line_type, indent, data = self.here()

        if (
            line_type == 'FigLine' and
            indent == current_indent
        ):
            fig = {
                'tag': 'Fig',
                'attr': {
                    'src': data['body'],
                },
                'children': [],
            }
            self.forward()

        return fig

    def process_list(self, current_indent):
        self.skip_blank_lines()
        list_ = None
        list_children = []

        line_type, indent, data = self.here()
        next_line_type, next_indent, next_data = self.next()
        body_split = split(data['body']) if len(data) else ['']

        if (
            line_type == 'DefaultLine' and
            indent == current_indent and
            body_split[0][len(body_split[0]) - 1] not in ('編', '章', '節', '款', '目', '条', '附', '付', '則')
        ):
            list_ = {
                'tag': 'List',
                'attr': {},
                'children': [{
                    'tag': 'ListSentence',
                    'attr': {},
                    'children': get_sentences(data['body']),
                }],
            }
            self.forward()

        return list_

    def process_suppl_provision(self):
        self.skip_blank_lines()
        suppl_provision = None
        suppl_provision_children = []

        line_type, indent, data = self.here()
        next_line_type, next_indent, next_data = self.next()
        match = re_SupplProvisionLabel.match(data['body'])
        titlejoin = re.sub(r'\s+', '', match.group(1)) if match else ''

        if (
            line_type == 'DefaultLine' and
            indent != 0 and
            match and
            (
                titlejoin.startswith('付則') or
                titlejoin.startswith('附則')
            )
        ):
            suppl_provision_attr = {}
            if match.group(3):
                suppl_provision_attr['Extract'] = 'true'
            if match.group(2):
                suppl_provision_attr['AmendLawNum'] = match.group(2)
            suppl_provision_children.append({
                'tag': 'SupplProvisionLabel',
                'attr': {},
                'children': [match.group(1)],
            })
            self.forward()

            articles_or_paragraph_items_processed = False

            while self.continuing():
                paragraph_items = self.process_paragraph_items()
                if len(paragraph_items):
                    suppl_provision_children.extend(paragraph_items)
                    articles_or_paragraph_items_processed = True
                else:
                    articles = self.process_articles()
                    if len(articles):
                        suppl_provision_children.extend(articles)
                        articles_or_paragraph_items_processed = True
                    else:
                        break

            line_type, indent, data = self.here()
            if (
                not articles_or_paragraph_items_processed and
                line_type == 'DefaultLine' and
                indent == 0
            ):

                suppl_provision_children.append({
                    'tag': 'Paragraph',
                    'attr': {},
                    'children': [
                        {
                            'tag': 'ParagraphNum',
                            'attr': {},
                            'children': [],
                        },
                        {
                            'tag': 'ParagraphSentence',
                            'attr': {},
                            'children': get_sentences(data['body']),
                        },
                    ],
                })
                self.forward()

            if len(suppl_provision_children):
                suppl_provision = {
                    'tag': 'SupplProvision',
                    'attr': suppl_provision_attr,
                    'children': suppl_provision_children,
                }

        return suppl_provision

    def process_appdx_table(self):
        self.skip_blank_lines()
        appdx_table = None
        appdx_table_children = []

        line_type, indent, data = self.here()
        next_line_type, next_indent, next_data = self.next()
        match = re_AppdxTableLabel.match(data['body'])
        titlejoin = re.sub(r'\s+', '', match.group(1)) if match else ''

        if (
            line_type == 'DefaultLine' and
            indent == 0 and
            match and
            titlejoin.startswith('別表')
        ):
            appdx_table_children.append({
                'tag': 'AppdxTableTitle',
                'attr': {},
                'children': [match.group(1)],
            })
            if match.group(2):
                appdx_table_children.append({
                    'tag': 'RelatedArticleNum',
                    'attr': {},
                    'children': [match.group(2)],
                })
            self.forward()

            current_indent = indent

            paragraph_items = self.process_paragraph_items(current_indent + 1)
            if len(paragraph_items):
                appdx_table_children.extend(paragraph_items)

            table_struct = self.process_table_struct(current_indent + 1)
            if table_struct:
                appdx_table_children.append(table_struct)

            appdx_table = {
                'tag': 'AppdxTable',
                'attr': {},
                'children': appdx_table_children,
            }

        return appdx_table

    def process_appdx_style(self):
        self.skip_blank_lines()
        appdx_style = None
        appdx_style_children = []
        style_struct_children = []

        line_type, indent, data = self.here()
        match = re_AppdxStyleLabel.match(data['body'])
        body_split = re.split(r'\s+', data['body'], maxsplit=1) if len(data) else ['']

        if (
            line_type == 'DefaultLine' and
            indent == 0 and
            body_split[0].startswith('様式')
        ):
            appdx_style_children.append({
                'tag': 'AppdxStyleTitle',
                'attr': {},
                'children': [match.group(1)],
            })
            if match.group(2):
                appdx_style_children.append({
                    'tag': 'RelatedArticleNum',
                    'attr': {},
                    'children': [match.group(2)],
                })
            self.forward()

            current_indent = indent

            remarks = self.process_remarks(current_indent + 1)
            if remarks:
                style_struct_children.append(remarks)

            fig = self.process_fig(current_indent + 1)
            if fig:
                style = {
                    'tag': 'Style',
                    'attr': {},
                    'children': [fig],
                }
                style_struct_children.append(style)

            remarks = self.process_remarks(current_indent + 1)
            if remarks:
                style_struct_children.append(remarks)

            if len(style_struct_children):
                appdx_style_children.append({
                    'tag': 'StyleStruct',
                    'attr': {},
                    'children': style_struct_children,
                })

            appdx_style = {
                'tag': 'AppdxStyle',
                'attr': {},
                'children': appdx_style_children,
            }

        return appdx_style

    def process_remarks(self, current_indent):
        self.skip_blank_lines()
        remarks = None
        remarks_children = []

        line_type, indent, data = self.here()
        body_split = re.split(r'\s+', data['body'], maxsplit=1) if len(data) else ['']

        if (
            line_type == 'DefaultLine' and
            indent == current_indent and
            body_split[0].startswith('備考')
        ):
            remarks_children.append({
                'tag': 'RemarksLabel',
                'attr': {},
                'children': [body_split[0]],
            })

            if len(body_split) > 1:
                remarks_children.extend(get_sentences(body_split[1]))

            self.forward()

            paragraph_items = self.process_paragraph_items(current_indent + 1)
            if len(paragraph_items):
                remarks_children.extend(paragraph_items)

            remarks = {
                'tag': 'Remarks',
                'attr': {},
                'children': remarks_children,
            }

        return remarks

    def process_law(self):
        self.skip_blank_lines()
        law_children = []
        law_body_children = []
        law_num = None

        law_title, law_num = self.process_title()
        if law_num:
            law_children.append(law_num)
        if law_title:
            law_body_children.append(law_title)

        while self.continuing():
            enact_statement = self.process_enact_statement()
            if enact_statement:
                law_body_children.append(enact_statement)
                continue
            break

        toc = self.process_toc()
        if toc:
            law_body_children.append(toc)

        main_provision = self.process_main_provision()
        if main_provision:
            law_body_children.append(main_provision)

        while self.continuing():
            suppl_provision = self.process_suppl_provision()
            if suppl_provision:
                law_body_children.append(suppl_provision)
                continue
            appdx_table = self.process_appdx_table()
            if appdx_table:
                law_body_children.append(appdx_table)
                continue
            appdx_style = self.process_appdx_style()
            if appdx_style:
                law_body_children.append(appdx_style)
                continue
            break

        law_children.append({
            'tag': 'LawBody',
            'attr': {},
            'children': law_body_children,
        })

        return {
            'tag': 'Law',
            'attr': {},
            'children': law_children,
        }

def parse_lawtext(lawtext):
    lines = re.split(r'\r\n', lawtext)
    if len(lines) <= 1:
        lines = re.split(r'\n', lawtext)

    try:
        lexed_lines = lex(lines)
    except LexerError as error:
        lineno = error.lineno
        raise LawtextParseError(error.message, lineno, lines)

    parser = Parser(lexed_lines)

    try:
        law = parser.process_law()
    except ParserError as error:
        lineno = error.lineno
        raise LawtextParseError(error.message, lineno, lines)
    except LawtextParseError as e:
        raise e
    except Exception as e:
        raise LawtextParseError('この行の処理中にエラーが発生しました。', parser.lineno, lines)

    if parser.continuing():
        lineno = parser.lineno
        raise LawtextParseError('この行の種類が判別できません。', lineno, lines)

    return law



import re
# import unicodedata

re_JpNum = re.compile(r'^(?P<s1000>(?P<d1000>\S*)千)?(?P<s100>(?P<d100>\S*)百)?(?P<s10>(?P<d10>\S*)十)?(?P<d1>\S*)?$')
jpnum_digits = {
    '一': 1, '二': 2, '三': 3, '四': 4, '五': 5,
    '六': 6, '七': 7, '八': 8, '九': 9,
}

def parse_jpnum(text):
    match = re_JpNum.match(text)
    if match:
        d1000 = jpnum_digits.get(match.group(2), 1 if match.group(1) else 0)
        d100  = jpnum_digits.get(match.group(4), 1 if match.group(3) else 0)
        d10   = jpnum_digits.get(match.group(6), 1 if match.group(5) else 0)
        d1    = jpnum_digits.get(match.group(7), 0)
        return str(d1000 * 1000 + d100 * 100 + d10 * 10 + d1)
    return None

def parse_romannum(text):
    num = 0
    for i, char in enumerate(text):
        if char in ('i', 'I', 'ｉ', 'Ｉ'):
            if (
                i + 1 < len(text) and
                text[i + 1] in ('x', 'X', 'ｘ', 'Ｘ')
            ):
                num -= 1
            else:
                num += 1
        if char in ('x', 'X', 'ｘ', 'Ｘ'):
            num += 10
    return num

eras = {
    '明治': 'Meiji', '大正': 'Taisho',
    '昭和': 'Showa', '平成': 'Heisei',
}

re_LawTypes = [
    (re.compile(r'^法律$'), 'Act'),
    (re.compile(r'^政令$'), 'CabinetOrder'),
    (re.compile(r'^勅令$'), 'ImperialOrder'),
    (re.compile(r'^\S*[^政勅]令$'), 'MinisterialOrdinance'),
    (re.compile(r'^\S*規則$'), 'Rule'),
]

re_LawNum = re.compile(r'(?P<era>\S+?)(?P<year>[一二三四五六七八九十]+)年(?P<law_type>\S+?)第(?P<num>[一二三四五六七八九十百千]+)号')

def decorate_law(el):
    el['attr']['Lang'] = 'ja'
    for subel in el['children']:
        if (
            subel['tag'] == 'LawNum' and
            len(subel['children'])
        ):
            law_num = subel['children'][0]
            match = re_LawNum.match(law_num)
            if match:
                era = eras.get(match.group(1))
                if era:
                    el['attr']['Era'] = era

                year = parse_jpnum(match.group(2))
                if year:
                    el['attr']['Year'] = str(year)

                for re_LawType, law_type in re_LawTypes:
                    law_type_match = re_LawType.match(match.group(3))
                    if law_type_match:
                        el['attr']['LawType'] = law_type
                        break

                num = parse_jpnum(match.group(4))
                if num:
                    el['attr']['Num'] = num

WIDE_NUMS = {
    '０': '0', '１': '1', '２': '2', '３': '3', '４': '4',
    '５': '5', '６': '6', '７': '7', '８': '8', '９': '9',
}

def replace_wide_num(text):
    ret = text
    for wide, narrow in WIDE_NUMS.items():
        ret = ret.replace(wide, narrow)
    return ret

re_NamedNum = re.compile(r'^(?P<circle>○?)第?(?P<num>[一二三四五六七八九十百千]+)\S*?(?P<branch>の\S+)?$')
IROHA_CHARS = 'イロハニホヘトチリヌルヲワカヨタレソツネナラムウヰノオクヤマケフコエテアサキユメミシヱヒモセスン'
re_ItemNum = re.compile(r'^\D*(?P<num>\d+)\D*$')

def parse_named_num(text):
    nums_group = []

    for subtext in text.replace('及び', '、').replace('から', '、').replace('まで', '').replace('～', '、').replace('・', '、').split('、'):

        match = re_NamedNum.match(subtext)
        if match:
            nums = [parse_jpnum(match.group(2))]
            if match.group(3):
                nums.extend([
                    parse_jpnum(b)
                    for b in match.group(3).split('の')
                    if b
                ])
            nums_group.append('_'.join(map(str, nums)))
            continue

        iroha_char_detected = False

        for i, char in enumerate(IROHA_CHARS):
            if char in subtext:
                nums_group.append(str(i + 1))
                iroha_char_detected = True
                break

        if not iroha_char_detected:
            subtext = replace_wide_num(subtext)
            match = re_ItemNum.match(subtext)
            if match:
                nums_group.append(match.group(1))
                continue

        roman_num = parse_romannum(subtext)
        if roman_num:
            nums_group.append(str(roman_num))

    return ':'.join(nums_group)

# re_ParagraphNum = re.compile(r'^(?P<circle>○?)(?P<num>\d+)$')

# def parse_paragraph_num(text):
#     text = unicodedata.normalize('NFKC', text)
#     match = re_ParagraphNum.match(text)
#     if match:
#         return match.group('circle'), int(match.group('num'))


# def parse_item_num(text):
#     num = parse_named_num(text)
#     if num:
#         return num
#     for i, char in enumerate(IROHA_CHARS):
#         if char in text:
#             return i + 1
#     text = unicodedata.normalize('NFKC', text)
#     match = re_ItemNum.match(text)
#     if match:
#         return int(match.group('num'))

def decorate_toc_article_group(el):
    el['attr']['Delete'] = 'false'
    for subel in el['children']:
        if (
            subel['tag'] in (
                'PartTitle', 'ChapterTitle', 'SectionTitle',
                'SubsectionTitle', 'DivisionTitle', 'ArticleTitle',
            ) and
            len(subel['children'])
        ):
            body = subel['children'][0]
            num = parse_named_num(body.split()[0])
            if num:
                el['attr']['Num'] = num

def decorate_article_group(el):
    el['attr']['Delete'] = 'false'
    el['attr']['Hide'] = 'false'
    for subel in el['children']:
        if (
            subel['tag'] in (
                'PartTitle', 'ChapterTitle', 'SectionTitle',
                'SubsectionTitle', 'DivisionTitle', 'ArticleTitle',
            ) and
            len(subel['children'])
        ):
            body = subel['children'][0]
            num = parse_named_num(body.split()[0])
            if num:
                el['attr']['Num'] = num

def decorate_article(el):
    el['attr']['Delete'] = 'false'
    el['attr']['Hide'] = 'false'
    for subel in el['children']:
        if (
            subel['tag'] == 'ArticleTitle' and
            len(subel['children'])
        ):
            body = subel['children'][0]
            num = parse_named_num(body.split()[0])
            if num:
                el['attr']['Num'] = num

def decorate_paragraph(el):
    el['attr']['Hide'] = 'false'
    el['attr']['OldStyle'] = 'false'
    for subel in el['children']:
        if (
            subel['tag'] == 'ParagraphNum'
        ):
            if len(subel['children']):
                paragraph_num = subel['children'][0]
                num = parse_named_num(paragraph_num)
            else:
                num = '1'
            if num:
                el['attr']['Num'] = num

def decorate_item(el):
    el['attr']['Delete'] = 'false'
    el['attr']['Hide'] = 'false'
    for subel in el['children']:
        if (
            subel['tag'] in (
                'ItemTitle',
                'Subitem1Title', 'Subitem2Title', 'Subitem3Title',
                'Subitem4Title', 'Subitem5Title', 'Subitem6Title',
                'Subitem7Title', 'Subitem8Title', 'Subitem9Title',
                'Subitem10Title',
            )
        ):
            body = subel['children'][0]
            num = parse_named_num(body.split()[0])
            if num:
                el['attr']['Num'] = num

def decorate_column_sentence_group(el):
    column_sentences = []
    for subel in el['children']:
        if (
            subel['tag'] in ('Column', 'Sentence')
        ):
            column_sentences.append(subel)

    proviso_nums = []
    for i, subel in enumerate(column_sentences):
        if(len(column_sentences) > 1):
            subel['attr']['Num'] = str(i + 1)
        if (
            subel['tag'] == 'Column'
        ):
            subel['attr']['LineBreak'] = 'false'
        if (
            subel['tag'] == 'Sentence' and
            len(subel['children']) and
            (
                subel['children'][0].startswith('ただし、') or
                subel['children'][0].startswith('但し、')
            )
        ):
            proviso_nums.append(i)

    if len(proviso_nums):
        for i, subel in enumerate(column_sentences):
            subel['attr']['Function'] = 'proviso' if i in proviso_nums else 'main'

def decorate_sentence(el):
    el['attr']['WritingMode'] = 'vertical'

def decorate_table(el):
    el['attr']['WritingMode'] = 'vertical'

def decorate_table_column(el):
    el['attr']['BorderTop'] = 'solid'
    el['attr']['BorderRight'] = 'solid'
    el['attr']['BorderBottom'] = 'solid'
    el['attr']['BorderLeft'] = 'solid'

def appdx_table_title(el):
    el['attr']['WritingMode'] = 'vertical'

def decorate(el):

    if el['tag'] == 'Law':
        decorate_law(el)

    elif el['tag'] in (
        'TOCPart', 'TOCChapter', 'TOCSection',
        'TOCSubsection', 'TOCDivision', 'TOCArticle',
    ):
        decorate_toc_article_group(el)

    elif el['tag'] in (
        'Part', 'Chapter', 'Section',
        'Subsection', 'Division',
    ):
        decorate_article_group(el)

    elif el['tag'] == 'Article':
        decorate_article(el)

    elif el['tag'] == 'Paragraph':
        decorate_paragraph(el)

    elif el['tag'] == 'Sentence':
        decorate_sentence(el)

    elif el['tag'] in (
        'Item',
        'Subitem1', 'Subitem2', 'Subitem3', 'Subitem4', 'Subitem5',
        'Subitem6', 'Subitem7', 'Subitem8', 'Subitem9', 'Subitem10',
    ):
        decorate_item(el)

    elif el['tag'] in (
        'ParagraphSentence', 'ItemSentence',
        'Subitem1Sentence', 'Subitem2Sentence', 'Subitem3Sentence',
        'Subitem4Sentence', 'Subitem5Sentence', 'Subitem6Sentence',
        'Subitem7Sentence', 'Subitem8Sentence', 'Subitem9Sentence',
        'Subitem10Sentence',
        'Column',
    ):
        decorate_column_sentence_group(el)

    elif el['tag'] == 'Table':
        decorate_table(el)

    elif el['tag'] == 'TableColumn':
        decorate_table_column(el)

    elif el['tag'] == 'AppdxTableTitle':
        appdx_table_title(el)

    for subel in el['children']:
        if isinstance(subel, str):
            continue
        decorate(subel)


from pathlib import Path
from io import BytesIO
from zipfile import ZipFile, ZIP_DEFLATED

from jinja2 import Environment, FileSystemLoader

env = Environment(loader=FileSystemLoader(str(Path(__file__).parent.resolve() / 'templates'), encoding='utf8'))



def element_to_json(el):
    children = []
    text = (el.text or '').strip()
    if text:
        children.append(text)
    for subel in el:
        children.append(element_to_json(subel))
        tail = (subel.tail or '').strip()
        if tail:
            children.append(tail)
    return {
        'tag': el.tag,
        'attr': el.attrib,
        'children': children,
    }

def render_htmlfragment(law):
    template = env.get_template('htmlfragment.html')
    return template.render({'law': law, 'context': Context()})



def render_html(law):
    template = env.get_template('html.html')
    style = (Path(__file__).parent.resolve() / 'static/law.css').read_text(encoding='utf-8')
    return template.render({'law': law, 'style': style, 'context': Context()})

def render_xml(law):
    template = env.get_template('xml.xml')
    return template.render({'law': law})

def render_lawtext(law):
    template = env.get_template('lawtext.j2')
    return template.render({'law': law, 'context': Context()})



def restructure_table(table):
    new_table_children = []
    rowspan_state = {}
    colspan_value = {}
    for row in table['children']:
        if row['tag'] != 'TableRow':
            new_table_children.append(row)
            continue
        new_row_children = []
        c = 0
        ci = 0
        while True:
            rss = rowspan_state.get(c, 0)

            if rss:
                colspan = colspan_value.get(c, 0)
                new_row_children.append({
                    'tag': 'TableColumnMerged',
                    'attr': {
                        'colspan': colspan,
                    } if colspan else {},
                    'children': [],
                })
                rowspan_state[c] = rss - 1
                if colspan:
                    c += colspan - 1
                c += 1
                continue

            if ci >= len(row['children']):
                break

            column = row['children'][ci]
            new_row_children.append(column)
            if column['tag'] != 'TableColumn':
                ci += 1
                continue

            colspan = int(column['attr'].get('colspan', 0))
            if 'rowspan' in column['attr']:
                rowspan = int(column['attr']['rowspan'])
                rowspan_state[c] = rowspan - 1
                colspan_value[c] = colspan
                if colspan:
                    c += colspan - 1

            c += 1
            ci += 1

        new_table_children.append({
            'tag': row['tag'],
            'attr': row['attr'],
            'children': new_row_children,
        })

    ret = {
        'tag': table['tag'],
        'attr': table['attr'],
        'children': new_table_children,
    }

    return ret



class Context:
    def __init__(self):
        self.data = {}
    def get(self, key):
        return self.data.get(key)
    def set(self, key, value):
        self.data[key] = value
        return ''



def render_docx(law):
    t_content_types = env.get_template('docx/[Content_Types].xml')
    t_rels = env.get_template('docx/_rels/.rels')
    t_document_rels = env.get_template('docx/word/_rels/document.xml.rels')
    t_document = env.get_template('docx/word/document.xml')
    t_styles = env.get_template('docx/word/styles.xml')

    s_content_types = t_content_types.render()
    s_rels = t_rels.render()
    s_document_rels = t_document_rels.render()
    s_document = t_document.render({
        'law': law,
        'restructure_table': restructure_table,
        'print': print,
        'context': Context(),
    })
    s_styles = t_styles.render()

    buffer = BytesIO()
    with ZipFile(buffer, mode='w', compression=ZIP_DEFLATED) as zipfile:
        zipfile.writestr('[Content_Types].xml', s_content_types)
        zipfile.writestr('_rels/.rels', s_rels)
        zipfile.writestr('word/_rels/document.xml.rels', s_document_rels)
        zipfile.writestr('word/document.xml', s_document)
        zipfile.writestr('word/styles.xml', s_styles)

    return buffer.getvalue()



import io
import re
import sys
import unittest
from abc import ABCMeta
import time

import requests

import requests_cache

requests_cache.install_cache()

_web_law_cache = {}
_ref_text_cache = {}


class AbstractTest(unittest.TestCase, metaclass=ABCMeta):

    LAWS = [
        dict(zip(
            ['law_id', 'law_num', 'law_name'],
            values,
        ))
        for values in [
            ('359AC0000000086', '昭和五十九年法律第八十六号', '電気通信事業法'),
            ('322AC0000000067', '昭和二十二年法律第六十七号', '地方自治法'),
            ('325AC0000000131', '昭和二十五年法律第百三十一号', '電波法'),
            ('425AC0000000027', '平成二十五年法律第二十七号', '行政手続における特定の個人を識別するための番号の利用等に関する法律'),
            ('414AC0000000153', '平成十四年法律第百五十三号', '電子署名等に係る地方公共団体情報システム機構の認証業務に関する法律'),
            ('405AC0000000088', '平成五年法律第八十八号', '行政手続法'),
            ('406CO0000000265', '平成六年政令第二百六十五号', '行政手続法施行令'),
            ('428M60000008031', '平成二十八年総務省令第三十一号', '第二種指定電気通信設備接続料規則'),
            ('426M60000002044', '平成二十六年内閣府令第四十四号', '子ども・子育て支援法施行規則'),
            ('346AC0000000073', '昭和四十六年法律第七十三号', '児童手当法'),
            ('129AC0000000089', '明治二十九年法律第八十九号', '民法'),
        ]
    ]

    def _compare_text(self, lines_1, lines_2, lines, law_num):
        conc1 = ''.join(line[2:].rstrip() for line in lines_1)
        conc2 = ''.join(line[2:].rstrip() for line in lines_2)
        if conc1 == conc2:
            return
        if (len(lines) == 4 and
            lines[0] == '*** 1 ****' and
            lines[1] == f'- {law_num}' and
            lines[2] == '--- 0 ----' and
            lines[3] == '***************'):
            return
        if (len(lines) == 4 and
            lines[0] == '*** 2 ****' and
            lines[1] == '--- 2 ----' and
            lines[2] == f'+ （{law_num}）' and
            lines[3] == '***************'):
            return
        yield from lines

    def _compare_xml(self, lines_1, lines_2, lines, law_num, filename1='reference', filename2='output'):
        conc1 = '\n'.join(line[2:] for line in lines_1)
        conc2 = '\n'.join(line[2:] for line in lines_2)
        if conc1 == conc2:
            return

        conc1 = conc1.replace(' Year="0', ' Year="')
        conc1 = conc1.replace(' Extract="false"', '')
        conc1 = conc1.replace(' Delete="false"', '')
        conc1 = conc1.replace(' Hide="false"', '')
        conc1 = conc1.replace(' OldStyle="false"', '')
        conc1 = conc1.replace(' WritingMode="vertical"', '')
        conc1 = conc1.replace(' BorderTop="solid"', '')
        conc1 = conc1.replace(' BorderRight="solid"', '')
        conc1 = conc1.replace(' BorderBottom="solid"', '')
        conc1 = conc1.replace(' BorderLeft="solid"', '')
        conc1 = conc1.replace(' LineBreak="false"', '')

        conc2 = conc2.replace(' Year="0', ' Year="')
        conc2 = conc2.replace(' Extract="false"', '')
        conc2 = conc2.replace(' Delete="false"', '')
        conc2 = conc2.replace(' Hide="false"', '')
        conc2 = conc2.replace(' OldStyle="false"', '')
        conc2 = conc2.replace(' WritingMode="vertical"', '')
        conc2 = conc2.replace(' BorderTop="solid"', '')
        conc2 = conc2.replace(' BorderRight="solid"', '')
        conc2 = conc2.replace(' BorderBottom="solid"', '')
        conc2 = conc2.replace(' BorderLeft="solid"', '')
        conc2 = conc2.replace(' LineBreak="false"', '')

        if conc1 == conc2:
            return

        conc1 = conc1.replace(' Function="main"', '')

        conc2 = conc2.replace(' Function="main"', '')

        if conc1 == conc2:
            return

        conc1 = conc1.replace(' Num="1"', '')

        conc2 = conc2.replace(' Num="1"', '')

        if conc1 == conc2:
            return

        conc1 = re.sub(r' Num="[\d:_]+?"', '', conc1)

        conc2 = re.sub(r' Num="[\d:_]+?"', '', conc2)

        if conc1 == conc2:
            out_text = '\n'.join(lines).rstrip()
            print(f'''\
###
### [Num] mismatch
###

*** {filename1}
--- {filename2}
***************

{out_text}
''')
            return

        conc1 = conc1.replace(' Function="proviso"', '')

        conc2 = conc2.replace(' Function="proviso"', '')

        if conc1 == conc2:
            out_text = '\n'.join(lines).rstrip()
            print(f'''\
###
### [Function] mismatch
###

*** {filename1}
--- {filename2}
***************

{out_text}
''')
            return

        conc1 = conc1.replace(' Type="Amend"', '')

        conc2 = conc2.replace(' Type="Amend"', '')

        if conc1 == conc2:
            out_text = '\n'.join(lines).rstrip()
            print(f'''\
###
### [Type] mismatch
###

*** {filename1}
--- {filename2}
***************

{out_text}
''')
            return

        conc2 = re.sub(
            r'<AppdxTableTitle(?P<appdx_table_attr>[^>]*?)>(?P<appdx_table_text>.*)</AppdxTableTitle>.*?<RelatedArticleNum>(?P<related_article_num_text>.*)</RelatedArticleNum>',
            r'<AppdxTableTitle\g<appdx_table_attr>>\g<appdx_table_text>\g<related_article_num_text></AppdxTableTitle>',
            conc2,
            flags=re.DOTALL,
        )

        if conc1 == conc2:
            out_text = '\n'.join(lines).rstrip()
            print(f'''\
###
### <RelatedArticleNum> not in **reference**
###

*** {filename1}
--- {filename2}
***************

{out_text}
''')
            return

        conc1 = re.sub(r'</Column>\s*<Column[^>]*?>', '　', conc1)
        conc1 = re.sub(r'。</Sentence>\s*<Sentence[^>]*?>', '。', conc1)
        conc1 = re.sub(r'</Sentence>\s*<Sentence[^>]*?>', '　', conc1)

        conc2 = re.sub(r'</Column>\s*<Column[^>]*?>', '　', conc2)
        conc2 = re.sub(r'。</Sentence>\s*<Sentence[^>]*?>', '。', conc2)
        conc2 = re.sub(r'</Sentence>\s*<Sentence[^>]*?>', '　', conc2)

        if conc1 == conc2:
            out_text = '\n'.join(lines).rstrip()
            print(f'''\
###
### <Column>'s and/or <Sentence>'s mismatch
###

*** {filename1}
--- {filename2}
***************

{out_text}
''')
            return

        yield from lines

    def _filter_diff_lines(self, lines, law_num, type_, filename1='reference', filename2='output'):
        sub_lines = []
        lines_1 = []
        lines_2 = []
        state = 0
        for line in lines:
            line = line.rstrip()
            sub_lines.append(line)
            if state == 0:
                if line.startswith('***'):
                    state = 1
                    continue
            elif state == 1:
                if line.startswith('---'):
                    state = 2
                    continue
                lines_1.append(line)
            elif state == 2:
                if line.startswith('******'):
                    state = 0
                    if type_ == 'text':
                        yield from self._compare_text(lines_1, lines_2, sub_lines, law_num)
                    elif type_ == 'xml':
                        yield from self._compare_xml(lines_1, lines_2, sub_lines, law_num, filename1=filename1, filename2=filename2)
                    else:
                        raise Exception
                    sub_lines = []
                    lines_1 = []
                    lines_2 = []
                    continue
                lines_2.append(line)
        if sub_lines:
            state = 0
            if type_ == 'text':
                yield from self._compare_text(lines_1, lines_2, sub_lines, law_num)
            elif type_ == 'xml':
                yield from self._compare_xml(lines_1, lines_2, sub_lines, law_num, filename1=filename1, filename2=filename2)
            else:
                raise Exception
            sub_lines = []
            lines_1 = []
            lines_2 = []

    def _flatten_html(self, html_raw):
        import lxml.html
        html_el = lxml.html.fromstring(html_raw)
        html_law_el = html_el.find('.//div[@class="LawBody"]')
        html_text = ''.join(html_law_el.itertext())
        return html_text

    def _request(self, url):
        import requests
        response = requests.get(url, headers={
            'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/62.0.3202.94 Safari/537.36',
        })
        return response.text

    def _get_web_law(self, law_num):
        if law_num in _web_law_cache:
            return _web_law_cache[law_num]
        from urllib.parse import quote
        from xml.etree import ElementTree as ET
        from html import unescape
        from lawtext.util import xml_to_json

        xml_url = f'http://elaws.e-gov.go.jp/api/1/lawdata/{quote(law_num)}'

        xml_raw = self._request(xml_url)
        xml_el = ET.fromstring(xml_raw)
        xml_law_data_text = unescape(xml_el.find('.//LawFullText').text)

        law = xml_to_json(xml_law_data_text)

        _web_law_cache[law_num] = law

        return law

    def _get_reference_text(self, law_id):
        if law_id in _ref_text_cache:
            return _ref_text_cache[law_id]
        import lxml.html

        html_url = f'http://elaws.e-gov.go.jp/search/elawsSearch/elaws_search/lsg0500/viewContents?lawId={law_id}'

        html_raw = self._request(html_url)
        html_el = lxml.html.fromstring(html_raw)
        html_law_el = html_el.find('.//div[@class="LawBody"]')
        html_text = '\n'.join(html_law_el.itertext())

        _ref_text_cache[law_id] = html_text

        return html_text

    def _get_diff_out_text(self, lines1, lines2, law_num, type_, filename1='reference', filename2='output'):
        from difflib import context_diff
        lines = context_diff(lines1, lines2, n={'text': 0, 'xml': 1}[type_])
        lines = self._filter_diff_lines(lines, law_num, type_, filename1=filename1, filename2=filename2)

        out_text = '\n'.join(lines)
        return out_text.rstrip()

    def _fail_diff_out_text(self, law_name, law_num, out_text, filename1='reference', filename2='output'):
        self.fail(f'''\
Mismatching Output:
{law_name}
（{law_num}）

*** {filename1}
--- {filename2}
***************

{out_text}''')


class TestRender(AbstractTest):
    def test_render_htmlfragment_with_elaws(self):
        import lxml.html
        from lawtext.render import render_htmlfragment
        for law in self.LAWS:
            law_num, law_id, law_name = law['law_num'], law['law_id'], law['law_name']

            with self.subTest(**law):

                print(f'\nSubtest {law_name}（{law_num}）', file=sys.stderr)

                # print('  Downloading web XML ...', file=sys.stderr)
                xml_law_data_text = self._get_web_law(law_num)
                # print('  Rendering HTML fragment ...', file=sys.stderr)
                rendered_raw = render_htmlfragment(xml_law_data_text)

                # print('  Extracting text from HTML fragment ...', file=sys.stderr)
                rendered_el = lxml.html.fromstring(rendered_raw)
                rendered_text = '\n'.join(rendered_el.itertext())

                self.assertNotEqual(rendered_text, '')

                rendered_text = re.sub(
                    r'\[.+?\]',
                    '[別画面で表示]',
                    rendered_text,
                )

                print('  Comparing ...', file=sys.stderr)
                out_text = self._get_diff_out_text(
                    self._get_reference_text(law_id).split(),
                    rendered_text.split(),
                    law_num,
                    type_='text',
                )

                if out_text:
                    self._fail_diff_out_text(law_name, law_num, out_text)

    def test_render_html_with_elaws(self):
        import lxml.html
        from lawtext.render import render_html
        from pathlib import Path

        out_test_dir = Path('out_test_parse')

        for law in self.LAWS:
            law_num, law_id, law_name = law['law_num'], law['law_id'], law['law_name']

            with self.subTest(**law):

                print(f'\nSubtest {law_name}（{law_num}）', file=sys.stderr)

                # print('  Downloading web XML ...', file=sys.stderr)
                xml_law_data_text = self._get_web_law(law_num)
                # print('  Rendering HTML ...', file=sys.stderr)
                rendered_raw = render_html(xml_law_data_text)
                rendered_out_path = (out_test_dir / f'rendered_{law_name}.html').resolve()
                print(f'    Writing "{str(rendered_out_path)}" ...', file=sys.stderr)
                rendered_out_path.write_text(rendered_raw, encoding='utf-8')

                # print('  Extracting text from HTML ...', file=sys.stderr)
                rendered_el = lxml.html.fromstring(rendered_raw)
                rendered_body_el = rendered_el.find('.//body')
                rendered_text = '\n'.join(rendered_body_el.itertext())

                self.assertNotEqual(rendered_text, '')

                rendered_text = re.sub(
                    r'\[.+?\]',
                    '[別画面で表示]',
                    rendered_text,
                )

                print('  Comparing ...', file=sys.stderr)
                out_text = self._get_diff_out_text(
                    self._get_reference_text(law_id).split(),
                    rendered_text.split(),
                    law_num,
                    type_='text',
                )

                if out_text:
                    self._fail_diff_out_text(law_name, law_num, out_text)

    def test_render_docx_with_elaws(self):
        from lawtext.render import render_docx
        from zipfile import ZipFile
        import xml.etree.ElementTree as ET
        from pathlib import Path

        out_test_dir = Path('out_test_parse')

        for law in self.LAWS:
            law_num, law_id, law_name = law['law_num'], law['law_id'], law['law_name']

            with self.subTest(**law):

                print(f'\nSubtest {law_name}（{law_num}）', file=sys.stderr)

                # print('  Downloading web XML ...', file=sys.stderr)
                xml_law_data_text = self._get_web_law(law_num)
                # print('  Rendering docx ...', file=sys.stderr)
                rendered_docx = render_docx(xml_law_data_text)
                rendered_out_path = (out_test_dir / f'rendered_{law_name}.docx').resolve()
                print(f'    Writing "{str(rendered_out_path)}" ...', file=sys.stderr)
                rendered_out_path.write_bytes(rendered_docx)

                # print('  Extracting text from docx ...', file=sys.stderr)
                with io.BytesIO(rendered_docx) as f:
                    with ZipFile(f) as zip_:
                        rendered_document = zip_.read('word/document.xml').decode()
                rendered_document = re.sub(
                    r'<(?P<slash>/?)w:(?P<tag>\S+?)(?P<attr>[^>]*?)>',
                    r'<\g<slash>w_\g<tag>\g<attr>>',
                    rendered_document,
                )
                rendered_el = ET.fromstring(rendered_document)
                for ruby in rendered_el.findall('.//w_ruby'):
                    rt = ruby.find('w_rt')
                    ruby.remove(rt)
                    ruby.append(rt)
                rendered_texts = [
                    subtext
                    for el in rendered_el.findall('.//w_t')
                    for subtext in el.text.split()
                ]
                rendered_text = '\n'.join(rendered_texts)
                rendered_text = re.sub(
                    r'（\s*(?P<body>\S+)\s*）',
                    r'（\g<body>）',
                    rendered_text,
                )
                rendered_text = re.sub(
                    r'\[.+?\]',
                    '[別画面で表示]',
                    rendered_text,
                )

                self.assertNotEqual(rendered_text, '')

                print('  Comparing ...', file=sys.stderr)
                out_text = self._get_diff_out_text(
                    self._get_reference_text(law_id).split(),
                    rendered_text.splitlines(),
                    law_num,
                    type_='text',
                )

                if out_text:
                    self._fail_diff_out_text(law_name, law_num, out_text)

    def test_render_lawtext_with_elaws(self):
        from lawtext.render import render_lawtext

        for law in self.LAWS:
            law_num, law_id, law_name = law['law_num'], law['law_id'], law['law_name']

            with self.subTest(**law):

                print(f'\nSubtest {law_name}（{law_num}）', file=sys.stderr)

                # print('  Downloading web XML ...', file=sys.stderr)
                xml_law_data_text = self._get_web_law(law_num)
                # print('  Rendering Lawtext ...', file=sys.stderr)
                rendered_text = render_lawtext(xml_law_data_text)

                # print('  Extracting text from Lawtext ...', file=sys.stderr)
                rendered_text = re.sub(
                    r'^\s*(?:\* )?- (?:\[\S+?\])*(.*)$',
                    r'\1',
                    rendered_text,
                    flags=re.M,
                )
                rendered_text = re.sub(
                    r'(?:</?Ruby>)|(?:</?Rt>)',
                    '',
                    rendered_text,
                    flags=re.M,
                )
                rendered_text = re.sub(
                    r'^(\s*)\.\.(?:\s*)figure::(?:\s*)(?:.+)$',
                    r'\1[別画面で表示]',
                    rendered_text,
                    flags=re.M,
                )

                self.assertNotEqual(rendered_text, '')

                print('  Comparing ...', file=sys.stderr)
                out_text = self._get_diff_out_text(
                    self._get_reference_text(law_id).split(),
                    rendered_text.split(),
                    law_num,
                    type_='text',
                )

                if out_text:
                    self._fail_diff_out_text(law_name, law_num, out_text)



class TestParse(AbstractTest):
    def test_parse_lawtext_with_elaws(self):
        from lawtext.render import render_lawtext, render_xml
        from lawtext.parse import parse_lawtext
        from lawtext.decorate import decorate
        from pathlib import Path
        from xml.dom import minidom

        out_test_dir = Path('out_test_parse')

        for law in self.LAWS:
            law_num, law_id, law_name = law['law_num'], law['law_id'], law['law_name']

            with self.subTest(**law):

                print(f'\nSubtest {law_name}（{law_num}）', file=sys.stderr)
                out_test_dir.mkdir(parents=True, exist_ok=True)

                # print('  Downloading web XML ...', file=sys.stderr)
                web_law = self._get_web_law(law_num)
                # print('  Parsing web XML ...', file=sys.stderr)
                web_xml = render_xml(web_law)
                web_outtext = minidom.parseString(web_xml).toprettyxml(indent='  ')
                web_out_path = (out_test_dir / f'web_{law_name}.xml').resolve()
                print(f'    Writing "{str(web_out_path)}" ...', file=sys.stderr)
                web_out_path.write_text(web_outtext, encoding='utf-8')

                # print('  Rendering lawtext ...', file=sys.stderr)
                lawtext = render_lawtext(web_law)
                rendered_out_path = (out_test_dir / f'rendered_{law_name}.law.txt').resolve()
                print(f'    Writing "{str(rendered_out_path)}" ...', file=sys.stderr)
                rendered_out_path.write_text(lawtext, encoding='utf-8')

                print('  Parsing lawtext ...', file=sys.stderr)

                t0 = time.time()
                parsed_law = parse_lawtext(lawtext)
                t1 = time.time()
                decorate(parsed_law)
                t2 = time.time()
                lines_count = len(lawtext.splitlines())

                print(f'    parse:    {(t1 - t0) * 1_000_000 / lines_count:3,.0f} μs/line  =  {(t1 - t0) * 1_000:5,.0f} ms / {lines_count:,} lines')
                print(f'    decorate: {(t2 - t1) * 1_000_000 / lines_count:3,.0f} μs/line  =  {(t2 - t1) * 1_000:5,.0f} ms / {lines_count:,} lines')

                parsed_xml = render_xml(parsed_law)
                parsed_outtext = minidom.parseString(parsed_xml).toprettyxml(indent="  ")
                parsed_out_path = (out_test_dir / f'parsed_{law_name}.xml').resolve()
                print(f'    Writing "{str(parsed_out_path)}" ...', file=sys.stderr)
                parsed_out_path.write_text(parsed_outtext, encoding='utf-8')

                print('  Comparing ...', file=sys.stderr)
                warning_outtext = None
                original_stdout = sys.stdout
                with io.StringIO() as buffer:
                    sys.stdout = buffer
                    out_text = self._get_diff_out_text(
                        web_outtext.splitlines(),
                        parsed_outtext.splitlines(),
                        law_num,
                        type_='xml',
                        filename1=str(web_out_path),
                        filename2=str(parsed_out_path),
                    )
                    warning_outtext = buffer.getvalue()
                    sys.stdout = original_stdout
                sys.stdout = original_stdout

                if warning_outtext:
                    warning_out_path = (out_test_dir / f'warning_{law_name}.diff').resolve()
                    print(f'    ###', file=sys.stderr)
                    print(f'    ### Some warning detected !!!', file=sys.stderr)
                    print(f'    ### Please see "{str(warning_out_path)}"', file=sys.stderr)
                    print(f'    ###', file=sys.stderr)
                    warning_out_path.write_text(warning_outtext, encoding='utf-8')

                if out_text:
                    self._fail_diff_out_text(
                        law_name, law_num, out_text,
                        filename1=str(web_out_path),
                        filename2=str(parsed_out_path),
                    )



from pathlib import Path

from docutils import nodes
from docutils.parsers.rst import Directive, directives
from docutils.utils.error_reporting import ErrorString

from lawtext.parse import parse_lawtext
from lawtext.render import render_htmlfragment


class LawTextDirective(Directive):
    has_content = True

    option_spec = {
        'file': directives.path,
        'encoding': directives.encoding,
    }

    def run(self):
        attributes = {
            'format': 'html',
        }
        encoding = self.options.get(
            'encoding', self.state.document.settings.input_encoding)

        if self.content:
            if 'file' in self.options:
                raise self.error(
                    '"%s" directive may not both specify an external file '
                    'and have content.' % self.name)
            lawtext = '\n'.join(self.content)

        elif 'file' in self.options:

            source_dir = Path(self.state.document.current_source).parent
            path = source_dir / self.options['file']
            try:
                lawtext = path.read_text(encoding)
                self.state.document.settings.record_dependencies.add(str(path))
            except IOError as error:
                raise self.severe(u'Problems with "%s" directive path:\n%s.'
                                  % (self.name, ErrorString(error)))
            except UnicodeError as error:
                raise self.severe(u'Problem with "%s" directive:\n%s'
                    % (self.name, ErrorString(error)))
            attributes['source'] = path

        law = parse_lawtext(lawtext)
        rendered = render_htmlfragment(law)
        raw_node = nodes.raw('', rendered, **attributes)
        (raw_node.source, raw_node.line) = self.state_machine.get_source_and_line(self.lineno)
        return [raw_node]



def setup(app):

    def builder_inited(app_):
        if app_.builder.name != 'html':
            return
        app_.builder.config.html_static_path.append(str(Path(__file__).resolve().parent / 'static'))
        app_.add_stylesheet('law.css')

    app.connect('builder-inited', builder_inited)
    app.add_directive('lawtext', LawTextDirective)



import xml.etree.ElementTree as ET

def xml_to_json(xml):
    return element_to_json(ET.fromstring(xml))

def element_to_json(el):
    children = []
    text = (el.text or '').strip()
    if text:
        children.append(text)
    for subel in el:
        children.append(element_to_json(subel))
        tail = (subel.tail or '').strip()
        if tail:
            children.append(tail)
    return {
        'tag': el.tag,
        'attr': el.attrib,
        'children': children,
    }

def json_to_element(obj):
    el = ET.Element(obj['tag'], attrib=obj['attr'])
    last_subel = None
    for child in obj['children']:
        if isinstance(child, str):
            if last_subel:
                last_subel.tail = (last_subel.tail or '') + child
            else:
                el.text = (el.text or '') + child
        else:
            subel = json_to_element(child)
            el.append(subel)
            last_subel = subel
    return el



from distutils.core import setup
from pathlib import Path

lawtext_dir = Path(__file__).parent / 'lawtext'

setup(
    name='lawtext',
    version='0.0.1',
    description='Plain text format for Japanese law',
    url='https://github.com/yamachig/lawtext',
    author='yamachi',
    license='MIT',
    packages=['lawtext'],
    package_data={'lawtext': [
        str(p.relative_to(lawtext_dir))
        for p in lawtext_dir.glob('**/*')
    ]},
    install_requires=['Jinja2', 'transcrypt'],
)
